/*
 * Copyright (c) 2016, National Institute of Informatics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package jp.gakunin.idp.attribute;

import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

import javax.annotation.Nonnull;
import javax.annotation.concurrent.NotThreadSafe;

import net.shibboleth.idp.attribute.IdPAttributeValue;
import net.shibboleth.idp.attribute.AttributeEncoder;

import net.shibboleth.utilities.java.support.annotation.constraint.NotEmpty;
import net.shibboleth.utilities.java.support.logic.Constraint;
import net.shibboleth.utilities.java.support.primitive.StringSupport;

import com.google.common.base.MoreObjects;
import com.google.common.base.Objects;

/**
 * Each attribute represents one piece of information about a user and has associated encoders used to turn that
 * information in to protocol-specific formats.
 *
 * Instances of {@link IdPAttribute} are compared using their IDs. That is, two attributes are considered the same if
 * they have the same ID, regardless of whether their display names, descriptions, values, or encoders are the same.
 */
@NotThreadSafe
public class UApproveJPIdPAttribute extends net.shibboleth.idp.attribute.IdPAttribute implements Comparable<net.shibboleth.idp.attribute.IdPAttribute>, Cloneable {

    /** Flag of optional attribute. */
    private boolean optional = false;

    /**
     * {@inheritDoc}
     */
    public UApproveJPIdPAttribute(@Nonnull @NotEmpty final String attributeId) {
        super(attributeId);
    }

    /**
     * Replaces the existing properties for this attribute with the given ones.
     * 
     * @param attribute Replaces the properties for this attribute with this. 
     */
    public void setIdPAttributeProperties(@Nonnull @NotEmpty final net.shibboleth.idp.attribute.IdPAttribute attribute) {
        setDisplayDescriptions(attribute.getDisplayDescriptions());
        setDisplayNames(attribute.getDisplayNames());
        setEncoders(attribute.getEncoders());
        setValues(attribute.getValues());
    }

    /**
     * Get the optional flag for this attribute
     *
     * @return boolean the optional flag for this attribute
     */
    @Nonnull @NotEmpty public boolean isOptional() {
        return optional;
    }

    /**
     * Get the mandatory flag for this attribute
     *
     * @return boolean the mandatory flag for this attribute
     */
    @Nonnull @NotEmpty public boolean isMandatory() {
        return !optional;
    }

    /**
     * Replaces the existing optional flag for this attribute with the given ones.
     *
     * @param boolean the optional flag for this attribute
     */
    public void setOptional(boolean flag) {
        optional = flag;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    @Nonnull public UApproveJPIdPAttribute clone() throws CloneNotSupportedException {
        final UApproveJPIdPAttribute clone = (UApproveJPIdPAttribute) super.clone();
        clone.setOptional(isOptional());
        return clone;
    }

    /** {@inheritDoc} */
    @Override
    public boolean equals(final Object obj) {
        if (this == obj) {
            return true;
        }

        if (obj == null) {
            return false;
        }

        if (!(obj instanceof net.shibboleth.idp.attribute.IdPAttribute) && !(obj instanceof UApproveJPIdPAttribute)) {
            return false;
        }

        net.shibboleth.idp.attribute.IdPAttribute other = (net.shibboleth.idp.attribute.IdPAttribute) obj;
        return java.util.Objects.equals(getId(), other.getId());
    }

    /** {@inheritDoc} */
    @Override
    @Nonnull public String toString() {
        return MoreObjects.toStringHelper(this).add("id", getId()).add("displayNames", getDisplayNames())
                .add("displayDescriptions", getDisplayDescriptions()).add("encoders", getEncoders()).add("values", getValues())
                .add("optional", isOptional())
                .toString();
    }
}
