/*
 * Copyright (c) 2016, National Institute of Informatics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package jp.gakunin.idp.attribute.filter.context;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.concurrent.NotThreadSafe;

import net.shibboleth.idp.attribute.IdPAttribute;
import net.shibboleth.idp.attribute.IdPAttributeValue;
import net.shibboleth.idp.attribute.filter.context.AttributeFilterContext;
import net.shibboleth.utilities.java.support.annotation.constraint.NonnullElements;
import net.shibboleth.utilities.java.support.annotation.constraint.NotEmpty;
import net.shibboleth.utilities.java.support.annotation.constraint.NullableElements;
import net.shibboleth.utilities.java.support.annotation.constraint.Unmodifiable;
import net.shibboleth.utilities.java.support.logic.Constraint;
import net.shibboleth.utilities.java.support.primitive.StringSupport;

import org.opensaml.messaging.context.BaseContext;

import com.google.common.collect.MapConstraints;

/**
 * Coordinates data between the filter implementation and the MatchFunctor implementations of uApproveJP.
 * 
 * <p>
 * This should be considered a private API limited to plugin implementations.
 * </p>
 */
@NotThreadSafe
public final class AttributeFilterOptionalContext extends BaseContext {

    /** Values, for a given attribute, that are optional to be released. */
    private final Map<String, Set<IdPAttributeValue>> optionalValues;

    /** Constructor. */
    public AttributeFilterOptionalContext() {
        optionalValues =
                MapConstraints.constrainedMap(new HashMap<String, Set<IdPAttributeValue>>(), MapConstraints.notNull());
    }

    /**
     * Gets the collection of attribute values, indexed by ID, that are optional to be released.
     * 
     * @return collection of attribute values, indexed by ID, that are optional to be released,
     */
    @Nonnull @NonnullElements @Unmodifiable public
            Map<String, Set<IdPAttributeValue>> getOptionalIdPAttributeValues() {
        return Collections.unmodifiableMap(optionalValues);
    }

    /**
     * Adds a collection of attribute values that are optional to be released. Attempting to add values for an
     * attribute that is not a member of {@link AttributeFilterContext#getPrefilteredIdPAttributes()} will result in an
     * {@link IllegalArgumentException}. Attempting to add an attribute value that is not a member of
     * {@link IdPAttribute#getValues()} will result in an {@link IllegalArgumentException}.
     * 
     * @param attributeId ID of the attribute whose values are optional to be released
     * @param attributeValues values for the attribute that are optional to be released
     */
    public void addOptionalIdPAttributeValues(@Nonnull @NotEmpty String attributeId,
            @Nullable @NullableElements Collection<? extends IdPAttributeValue> attributeValues) {

        AttributeFilterContext parent = (AttributeFilterContext) getParent();
        final Map<String, IdPAttribute> prefilteredAttributes = parent.getPrefilteredIdPAttributes();
        String trimmedAttributeId =
                Constraint.isNotNull(StringSupport.trimOrNull(attributeId), "Attribute ID can not be null or empty");
        Constraint.isTrue(prefilteredAttributes.containsKey(trimmedAttributeId), "no attribute with ID "
                + trimmedAttributeId + " exists in the pre-filtered attribute set");

        if (attributeValues == null || attributeValues.isEmpty()) {
            return;
        }

        Set<IdPAttributeValue> optionalAttributeValues = optionalValues.get(trimmedAttributeId);
        if (optionalAttributeValues == null) {
            optionalAttributeValues = new HashSet<>();
            optionalValues.put(trimmedAttributeId, optionalAttributeValues);
        }

        for (IdPAttributeValue value : attributeValues) {
            if (value != null) {
                if (!prefilteredAttributes.get(trimmedAttributeId).getValues().contains(value)) {
                    throw new IllegalArgumentException("optional value is not a current value of attribute "
                            + trimmedAttributeId);
                }

                if (!optionalAttributeValues.contains(value)) {
                    optionalAttributeValues.add(value);
                }
            }
        }
    }

}
