/*
 * Copyright (c) 2016, National Institute of Informatics
 */
/*
 * Licensed to the University Corporation for Advanced Internet Development, 
 * Inc. (UCAID) under one or more contributor license agreements.  See the 
 * NOTICE file distributed with this work for additional information regarding
 * copyright ownership. The UCAID licenses this file to You under the Apache 
 * License, Version 2.0 (the "License"); you may not use this file except in 
 * compliance with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package jp.gakunin.idp.attribute.filter.impl;

import java.util.Collection;
import java.util.List;
import java.util.Map;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.concurrent.ThreadSafe;

import net.shibboleth.idp.attribute.IdPAttribute;
import jp.gakunin.idp.attribute.UApproveJPIdPAttribute;
import net.shibboleth.idp.attribute.filter.AttributeFilter;
import net.shibboleth.idp.attribute.filter.AttributeFilterException;
import net.shibboleth.idp.attribute.filter.AttributeFilterPolicy;
import net.shibboleth.idp.attribute.filter.context.AttributeFilterContext;
import net.shibboleth.idp.attribute.filter.context.AttributeFilterWorkContext;
import jp.gakunin.idp.attribute.filter.context.AttributeFilterOptionalContext;
import net.shibboleth.utilities.java.support.annotation.constraint.NotEmpty;
import net.shibboleth.utilities.java.support.annotation.constraint.NullableElements;
import net.shibboleth.utilities.java.support.component.ComponentSupport;
import net.shibboleth.utilities.java.support.logic.Constraint;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/** Service that filters out attributes and values based upon loaded policies. */
@ThreadSafe
public class AttributeFilterImpl extends net.shibboleth.idp.attribute.filter.impl.AttributeFilterImpl implements AttributeFilter {

    /** Class logger. */
    private final Logger log = LoggerFactory.getLogger(AttributeFilterImpl.class);

    /**
     * {@inheritDoc}
     */
    public AttributeFilterImpl(@Nonnull @NotEmpty String engineId,
            @Nullable @NullableElements final Collection<AttributeFilterPolicy> policies) {
        super(engineId, policies);
    }

    /**
     * {@inheritDoc}
     */
    @Override public void filterAttributes(@Nonnull final AttributeFilterContext filterContext)
            throws AttributeFilterException {
        ComponentSupport.ifNotInitializedThrowUninitializedComponentException(this);
        ComponentSupport.ifDestroyedThrowDestroyedComponentException(this);

        Constraint.isNotNull(filterContext, "Attribute filter context can not be null");
        Map<String, IdPAttribute> prefilteredAttributes = filterContext.getPrefilteredIdPAttributes();

        // Create work context to hold intermediate results.
        filterContext.getSubcontext(AttributeFilterWorkContext.class, true);

        // Create optional context to hold intermediate results.
        filterContext.getSubcontext(AttributeFilterOptionalContext.class, true);

        log.debug("Beginning process of filtering the following {} attributes: {}", new Object[] {
                prefilteredAttributes.size(), prefilteredAttributes.keySet(),});

        final List<AttributeFilterPolicy> policies = getFilterPolicies();
        for (AttributeFilterPolicy policy : policies) {
            policy.apply(filterContext);
        }

        UApproveJPIdPAttribute filteredAttribute;
        for (String attributeId : filterContext.getPrefilteredIdPAttributes().keySet()) {
            final Collection filteredAttributeValues = getFilteredValues(attributeId, filterContext);
            if (filteredAttributeValues != null && !filteredAttributeValues.isEmpty()) {
                try {
                    filteredAttribute = new UApproveJPIdPAttribute(attributeId);
                    filteredAttribute.setIdPAttributeProperties(prefilteredAttributes.get(attributeId).clone());
                } catch (CloneNotSupportedException e) {
                    throw new AttributeFilterException(e);
                }
                filteredAttribute.setValues(filteredAttributeValues);
                filteredAttribute.setOptional(getOptionalMode(attributeId, filterContext));
                filterContext.getFilteredIdPAttributes().put(filteredAttribute.getId(), filteredAttribute);
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    @Nullable protected Collection getFilteredValues(@Nonnull @NotEmpty final String attributeId,
            @Nonnull final AttributeFilterContext filterContext) {
        Constraint.isNotNull(attributeId, "attributeId can not be null");
        Constraint.isNotNull(filterContext, "filterContext can not be null");

        final AttributeFilterWorkContext filterWorkContext =
                filterContext.getSubcontext(AttributeFilterWorkContext.class, false);
        Constraint.isNotNull(filterWorkContext, "Attribute filter work context can not be null");

        final AttributeFilterOptionalContext filterOptionalContext =
                filterContext.getSubcontext(AttributeFilterOptionalContext.class, false);
        Constraint.isNotNull(filterWorkContext, "Attribute filter optional context can not be null");

        Collection filteredAttributeValues = filterWorkContext.getPermittedIdPAttributeValues().get(attributeId);

        if (filteredAttributeValues == null || filteredAttributeValues.isEmpty()) {
            filteredAttributeValues = filterOptionalContext.getOptionalIdPAttributeValues().get(attributeId);
            if (filteredAttributeValues == null || filteredAttributeValues.isEmpty()) {
                log.debug("Attribute filtering engine '{}': no policy permitted release of attribute {} values", getId(),
                        attributeId);
                return null;
            }
        }

        if (filterWorkContext.getDeniedAttributeValues().containsKey(attributeId)) {
            filteredAttributeValues.removeAll(filterWorkContext.getDeniedAttributeValues().get(attributeId));
        }

        if (filteredAttributeValues.isEmpty()) {
            log.debug("Attribute filtering engine '{}': deny policies filtered out all values for attribute '{}'",
                    getId(), attributeId);
        } else {
            log.debug("Attribute filtering engine '{}': {} values for attribute '{}' remained after filtering",
                    new Object[] {getId(), filteredAttributeValues.size(), attributeId,});
        }

        return filteredAttributeValues;
    }

    /**
     * Get the optional mode for the given attribute from the
     * (@link AttributeFilterOptionalContext#OptionalIdPAttributeValues()) and removes all
     * permitted/denied values give in the
     * {@link AttributeFilterWorkContext#getPermittedIdPAttributeValues()}
     * {@link AttributeFilterWorkContext#getDeniedAttributeValues()}.
     * 
     * @param attributeId ID of the attribute whose values are to be retrieved
     * @param filterContext current attribute filter context
     * 
     * @return boolean optional mode of attribute
     */
    @Nullable private boolean getOptionalMode(@Nonnull @NotEmpty final String attributeId,
            @Nonnull final AttributeFilterContext filterContext) {
        Constraint.isNotNull(attributeId, "attributeId can not be null");
        Constraint.isNotNull(filterContext, "filterContext can not be null");

        final AttributeFilterWorkContext filterWorkContext =
                filterContext.getSubcontext(AttributeFilterWorkContext.class, false);
        Constraint.isNotNull(filterWorkContext, "Attribute filter work context can not be null");

        final AttributeFilterOptionalContext filterOptionalContext =
                filterContext.getSubcontext(AttributeFilterOptionalContext.class, false);
        Constraint.isNotNull(filterWorkContext, "Attribute filter optional context can not be null");

        Collection filteredAttributeValues = filterWorkContext.getPermittedIdPAttributeValues().get(attributeId);

        if (filteredAttributeValues == null || filteredAttributeValues.isEmpty()) {
            filteredAttributeValues = filterOptionalContext.getOptionalIdPAttributeValues().get(attributeId);
            if (filteredAttributeValues == null || filteredAttributeValues.isEmpty()) {
                log.debug("Attribute filtering engine '{}': no policy permitted release of attribute {} values", getId(),
                        attributeId);
                return false;
            }
        } else {
            return false;
        }

        if (filterWorkContext.getDeniedAttributeValues().containsKey(attributeId)) {
            filteredAttributeValues.removeAll(filterWorkContext.getDeniedAttributeValues().get(attributeId));
        }

        if (filteredAttributeValues.isEmpty()) {
            return false;
        }

        return true;
    }

}
